/*
 * This file is a part of project QuickShop, the name is BlockListener.java
 *  Copyright (C) PotatoCraft Studio and contributors
 *
 *  This program is free software: you can redistribute it and/or modify it
 *  under the terms of the GNU General Public License as published by the
 *  Free Software Foundation, either version 3 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful, but WITHOUT
 *  ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
 *  FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public License
 *  for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program. If not, see <http://www.gnu.org/licenses/>.
 *
 */

package org.maxgamer.quickshop.listener;

import com.ghostchu.simplereloadlib.ReloadResult;
import com.ghostchu.simplereloadlib.ReloadStatus;
import io.papermc.lib.PaperLib;
import org.bukkit.GameMode;
import org.bukkit.Location;
import org.bukkit.Material;
import org.bukkit.block.Block;
import org.bukkit.block.BlockFace;
import org.bukkit.block.BlockState;
import org.bukkit.block.Sign;
import org.bukkit.block.data.BlockData;
import org.bukkit.block.data.Directional;
import org.bukkit.entity.Player;
import org.bukkit.event.EventHandler;
import org.bukkit.event.EventPriority;
import org.bukkit.event.block.BlockBreakEvent;
import org.bukkit.event.block.BlockPlaceEvent;
import org.bukkit.event.block.SignChangeEvent;
import org.bukkit.event.inventory.InventoryMoveItemEvent;
import org.jetbrains.annotations.NotNull;
import org.jetbrains.annotations.Nullable;
import org.maxgamer.quickshop.Cache;
import org.maxgamer.quickshop.QuickShop;
import org.maxgamer.quickshop.api.shop.Info;
import org.maxgamer.quickshop.api.shop.Shop;
import org.maxgamer.quickshop.api.shop.ShopAction;
import org.maxgamer.quickshop.util.Util;
import org.maxgamer.quickshop.util.logging.container.ShopRemoveLog;

/**
 * BlockListener to listening events about block events
 *
 * @author KaiNoMood, Ghost_chu, sandtechnology
 */
public class BlockListener extends AbstractProtectionListener {
    private boolean update_sign_when_inventory_moving;

    public BlockListener(@NotNull final QuickShop plugin, @Nullable final Cache cache) {
        super(plugin, cache);
        init();
    }

    private void init() {
        this.update_sign_when_inventory_moving = super.getPlugin().getConfig().getBoolean("shop.update-sign-when-inventory-moving", true);
    }

    /*
     * Removes chests when they're destroyed.
     */
    @EventHandler(priority = EventPriority.NORMAL, ignoreCancelled = true)
    public void onBreak(BlockBreakEvent e) {
        final Block b = e.getBlock();
        final Player p = e.getPlayer();
        // If the shop was a chest
        if (Util.canBeShop(b)) {
            final Shop shop = getShopPlayer(b.getLocation(), false);
            if (shop == null) {
                return;
            }
            // If they're either survival or the owner, they can break it
            if (p.getGameMode() == GameMode.CREATIVE && !p.getUniqueId().equals(shop.getOwner())) {
                // Check SuperTool
                if (p.getInventory().getItemInMainHand().getType() == Material.GOLDEN_AXE) {
                    if (getPlugin().getConfig().getBoolean("shop.disable-super-tool")) {
                        e.setCancelled(true);
                        plugin.text().of(p, "supertool-is-disabled").send();
                        return;
                    }
                    plugin.text().of(p, "break-shop-use-supertool").send();
                    return;
                }
                e.setCancelled(true);
                plugin.text().of(p, "no-creative-break", plugin.getPlatform().getItemTranslationKey(Material.GOLDEN_AXE)).send();
                return;
            }

            // Cancel their current menu... Doesnt cancel other's menu's.
            final Info action = super.getPlugin().getShopManager().getActions().get(p.getUniqueId());

            if (action != null) {
                action.setAction(ShopAction.CANCELLED);
            }
            plugin.logEvent(new ShopRemoveLog(e.getPlayer().getUniqueId(), "BlockBreak(player)", shop.saveToInfoStorage()));
            shop.delete();
            plugin.text().of(p, "success-removed-shop").send();
        } else if (Util.isWallSign(b.getType())) {
            BlockState state = PaperLib.getBlockState(b, false).getState();
            if (state instanceof Sign sign) {
                if (sign.getLine(0).equals(super.getPlugin().getConfig().getString("lockette.private"))
                        || sign.getLine(0).equals(super.getPlugin().getConfig().getString("lockette.more_users"))) {
                    // Ignore break lockette sign
                    return;
                }
            }

            final Shop shop = getShopNextTo(b.getLocation());

            if (shop == null) {
                return;
            }
            // If they're in creative and not the owner, don't let them
            // (accidents happen)
            if (p.getGameMode() == GameMode.CREATIVE && !p.getUniqueId().equals(shop.getOwner())) {
                // Check SuperTool
                if (p.getInventory().getItemInMainHand().getType() == Material.GOLDEN_AXE) {
                    if (getPlugin().getConfig().getBoolean("shop.disable-super-tool")) {
                        e.setCancelled(true);
                        plugin.text().of(p, "supertool-is-disabled").send();
                        return;
                    }
                    plugin.text().of(p, "break-shop-use-supertool").send();
                    plugin.logEvent(new ShopRemoveLog(e.getPlayer().getUniqueId(), "BlockBreak(player)", shop.saveToInfoStorage()));
                    shop.onUnload();
                    shop.delete();
                    return;
                }
                e.setCancelled(true);
                plugin.text().of(p, "no-creative-break",plugin.getPlatform().getItemTranslationKey(Material.GOLDEN_AXE)).send();
                return;
            }
            //Allow Shop owner break the shop sign(for sign replacement)
            if (getPlugin().getConfig().getBoolean("shop.allow-owner-break-shop-sign") && p.getUniqueId().equals(shop.getOwner())) {
                return;
            }
            Util.debugLog("Player cannot break the shop infomation sign.");
            e.setCancelled(true);
        }
    }

    /**
     * Gets the shop a sign is attached to
     *
     * @param loc The location of the sign
     * @return The shop
     */
    @Nullable
    private Shop getShopNextTo(@NotNull Location loc) {
        final Block b = Util.getAttached(loc.getBlock());
        // Util.getAttached(b)
        if (b == null) {
            return null;
        }
        return getShopPlayer(b.getLocation(), false);
    }

    @EventHandler(ignoreCancelled = true, priority = EventPriority.MONITOR)
    public void onInventoryMove(InventoryMoveItemEvent event) {
        if (!this.update_sign_when_inventory_moving) {
            return;
        }
        Location destination = event.getDestination().getLocation();
        Location source = event.getSource().getLocation();
        Shop destShop = null;
        Shop sourceShop = null;
        if (destination != null) {
            destination = Util.getBlockLocation(destination);
            destShop = getShopRedstone(destination, true);
        }
        if (source != null) {
            source = Util.getBlockLocation(source);
            sourceShop = getShopRedstone(source, true);
        }
        if (destShop != null) {
            super.getPlugin().getSignUpdateWatcher().scheduleSignUpdate(destShop);
        }
        if (sourceShop != null) {
            super.getPlugin().getSignUpdateWatcher().scheduleSignUpdate(sourceShop);
        }
    }

    /*
     * Listens for sign update to prevent other plugin or Purpur to edit the sign
     */
    @EventHandler(ignoreCancelled = true, priority = EventPriority.LOWEST)
    public void onSignUpdate(SignChangeEvent event) {
        Block posShopBlock = Util.getAttached(event.getBlock());
        if (posShopBlock == null) {
            return;
        }
        Shop shop = plugin.getShopManager().getShopIncludeAttached(posShopBlock.getLocation());
        if (shop == null) {
            return;
        }
        Player player = event.getPlayer();
        if (!shop.getModerator().isModerator(player.getUniqueId())) {
            plugin.text().of(player, "not-managed-shop").send();
            event.setCancelled(true);
        }
    }
    /*
     * Listens for chest placement, so a doublechest shop can't be created.
     */
    @EventHandler(ignoreCancelled = true, priority = EventPriority.LOWEST)
    public void onPlace(BlockPlaceEvent e) {

        final Material type = e.getBlock().getType();
        final Block placingBlock = e.getBlock();
        final Player player = e.getPlayer();

        if (type != Material.CHEST) {
            return;
        }
        Block chest = null;

        //Chest combine mechanic based checking
        if (player.isSneaking()) {
            Block blockAgainst = e.getBlockAgainst();
            if (blockAgainst.getType() == Material.CHEST && placingBlock.getFace(blockAgainst) != BlockFace.UP && placingBlock.getFace(blockAgainst) != BlockFace.DOWN && !Util.isDoubleChest(blockAgainst.getBlockData())) {
                chest = e.getBlockAgainst();
            } else {
                return;
            }
        } else {
            //Get all chest in vertical Location
            BlockFace placingChestFacing = ((Directional) (placingBlock.getBlockData())).getFacing();
            for (BlockFace face : Util.getVerticalFacing()) {
                //just check the right side and left side
                if (!face.equals(placingChestFacing) && !face.equals(placingChestFacing.getOppositeFace())) {
                    Block nearByBlock = placingBlock.getRelative(face);
                    BlockData nearByBlockData = nearByBlock.getBlockData();
                    if (nearByBlock.getType() == Material.CHEST
                            //non double chest
                            && !Util.isDoubleChest(nearByBlockData)
                            //same facing
                            && placingChestFacing == ((Directional) nearByBlockData).getFacing()) {
                        if (chest == null) {
                            chest = nearByBlock;
                        } else {
                            //when multiply chests competed, minecraft will always combine with right side
                            if (placingBlock.getFace(nearByBlock) == Util.getRightSide(placingChestFacing)) {
                                chest = nearByBlock;
                            }
                        }
                    }
                }
            }
        }
        if (chest == null) {
            return;
        }

        Shop shop = getShopPlayer(chest.getLocation(), false);
        if (shop != null) {
            if (!QuickShop.getPermissionManager().hasPermission(player, "quickshop.create.double")) {
                e.setCancelled(true);
                plugin.text().of(player, "no-double-chests").send();

            } else if (!shop.getModerator().isModerator(player.getUniqueId())) {
                e.setCancelled(true);
                plugin.text().of(player, "not-managed-shop").send();
            }
        }
    }

    /**
     * Callback for reloading
     *
     * @return Reloading success
     */
    @Override
    public ReloadResult reloadModule() {
        init();
        return ReloadResult.builder().status(ReloadStatus.SUCCESS).build();
    }
}
